import copy
import dataclasses
import logging
import typing
from .cache_manager import CachedOutputs
from .variable import Variable, EnvironmentVariable

logger = logging.getLogger(__name__)


class ExecutionLogger:
    """Log the execution of Tasks.

       Note that this object has longer lifetime than Context.
       One job can have multiple Context objects, but only one ExecutionLogger object.

       Task execution time is not logged in this class since we want to add a log entry before the task execution.
    """
    def __init__(self):
        self._logs = []

    def __deepcopy__(self, memo):
        return self  # Always return the same object

    def log(self, name, task_name, task_version, task_hash, config_dict, inputs_dict, cache_hit):
        record = {'name': name,
                  'task_name': task_name,
                  'task_version': task_version,
                  'task_hash': task_hash,
                  'config': config_dict,
                  'inputs': inputs_dict,
                  'cache_hit': cache_hit}
        self._logs.append(record)

    def get(self):
        return copy.deepcopy(self._logs)


class Context:
    """Manage variables for a single experiment."""
    def __init__(self, environment_variables: typing.Dict[str, str] = None, cache_manager=None):
        self._envs = copy.deepcopy(environment_variables or {})
        self._cache_manager = cache_manager
        self._outputs = {}
        self._execution_logger = ExecutionLogger()

    @property
    def is_cache_enabled(self):
        return bool(self._cache_manager)

    def add_outputs(self, name: str, outputs: typing.Union[dataclasses.dataclass, CachedOutputs]):
        """Add Task outputs to the context so that subsequent Tasks can consume them.

        Args:
            name (str): the name of the task.
            outputs (Outputs or CachedOutputs instance): the outputs of the task.
        """
        if name in self._outputs:
            logger.warning(f"Duplicated task name: {name}. The outputs are overwritten.")
        self._outputs[name] = outputs

    def get_outputs(self, output_name: str):
        """Get the outputs of previous tasks.
        Args:
            name (str): the name of the task
        Returns:
            Outputs dataclass. If the task had been skipped, returns CachedOutputs.
        """
        if output_name not in self._outputs:
            raise ValueError(f"Output {output_name} is not found.")
        return self._outputs[output_name]

    def add_environment_variable(self, name: str, value: str):
        self._envs[name] = value

    def get_environment_variable(self, name: str):
        if name not in self._envs:
            raise ValueError(f"Environment variable {name} is not found.")
        return self._envs[name]

    def resolve(self, value):
        """Recursively replace Variables with actual value.

        Supported containers are:
            - Dict
            - List
            - dataclasses.dataclass
        """
        if isinstance(value, dict):
            return {k: self.resolve(v) for k, v in value.items()}
        elif dataclasses.is_dataclass(value):
            return type(value)(**{field.name: self.resolve(getattr(value, field.name)) for field in dataclasses.fields(value)})
        elif isinstance(value, list):
            return [self.resolve(v) for v in value]
        elif isinstance(value, Variable):
            return value.resolve(self)
        else:
            return value

    def get_cached_outputs(self, task_name, task_version, task_hash: str, outputs_class) -> typing.Optional[CachedOutputs]:
        """Try to get cached outputs for the given task

        Args:
            task_name (str): Task name
            task_version (str): Task version
            task_hash (str): Task hash
            output_class (dataclass): The dataclass of the outputs

        Returns:
            CachedOutput instance if there is a cache. If not, returns None.
        """
        if not self._cache_manager:
            return None

        logger.debug(f"Trying to get cache for Task {task_name} version {task_version}. Hash: {task_hash}")
        return self._cache_manager.get_cache(task_name, task_version, task_hash, outputs_class)

    def add_cache_outputs(self, task_name, task_version, task_hash: str, outputs):
        """Save the task outputs to the cache storage."""
        if self._cache_manager:
            logger.debug(f"Uploading cache for Task {task_name} version {task_version}. Hash: {task_hash}")
            self._cache_manager.upload_cache(task_name, task_version, task_hash, outputs)

    def get_execution_logs(self):
        return self._execution_logger.get()

    def add_execution_log(self, name, task_name, task_version, task_hash, config_dict, inputs_dict, cache_hit):
        self._execution_logger.log(name, task_name, task_version, task_hash, self._resolve_envs(config_dict), self._resolve_envs(inputs_dict), cache_hit)

    def clone(self):
        return copy.deepcopy(self)

    def _resolve_envs(self, value):
        """Recursively replace Variables with actual value. Only replace environment variables.

        """
        if isinstance(value, dict):
            return {k: self._resolve_envs(v) for k, v in value.items()}
        elif dataclasses.is_dataclass(value):
            return {field.name: self._resolve_envs(getattr(value, field.name)) for field in dataclasses.fields(value)}
        elif isinstance(value, list):
            return [self._resolve_envs(v) for v in value]
        elif isinstance(value, EnvironmentVariable):
            try:
                resolved_value = value.resolve(self)
                if isinstance(resolved_value, (int, float, str, bool)):  # Only primitive types are allowed. We need to enforce str-only env variable in the future.
                    return resolved_value
            except ValueError:
                pass
            return str(value)
        elif isinstance(value, Variable):
            return str(value)  # We cannot afford to retrieve cached outputs.
        else:
            return value



import copy
import dataclasses
import typing
import unittest
from irisml.core.context import Context, ExecutionLogger
from irisml.core.variable import EnvironmentVariable, OutputVariable


class TestContext(unittest.TestCase):
    def test_clone(self):
        context = Context({'env': 'value'})
        new_context = context.clone()
        self.assertEqual(new_context.get_environment_variable('env'), 'value')

        new_context.add_environment_variable('env2', 'value2')
        new_context.add_outputs('out', None)

        with self.assertRaises(ValueError):
            context.get_environment_variable('env2')

        with self.assertRaises(ValueError):
            context.get_outputs('out')

    def test_resolve(self):
        context = Context()
        self.assertEqual(context.resolve(123), 123)
        self.assertEqual(context.resolve('123'), '123')
        self.assertEqual(context.resolve([1, 2, 3]), [1, 2, 3])
        self.assertEqual(context.resolve([1, '2', 3]), [1, '2', 3])

        with self.assertRaises(ValueError):
            context.resolve(EnvironmentVariable('$env.E'))

        context = Context({'E': 'v'})
        self.assertEqual(context.resolve(EnvironmentVariable('$env.E')), 'v')
        self.assertEqual(context.resolve([3, EnvironmentVariable('$env.E')]), [3, 'v'])

        class DummyOutput:
            value = 3

        context.add_outputs('test_out', DummyOutput())
        self.assertEqual(context.resolve([123, OutputVariable('$output.test_out.value')]), [123, 3])

    def test_resolve_nested_dataclass(self):
        @dataclasses.dataclass
        class ChildConfig:
            value: int

        @dataclasses.dataclass
        class Config:
            child: ChildConfig

        config = Config(ChildConfig(42))
        context = Context()
        self.assertEqual(context.resolve(config), config)

    def test_add_execution_log(self):
        context = Context({'ENV': 'value'})
        context.add_execution_log('name', 'task_name', '0.0.0', None, {'config': EnvironmentVariable('$env.ENV')}, {}, False)

        logs = context.get_execution_logs()
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0], {'name': 'name', 'task_name': 'task_name', 'task_version': '0.0.0', 'task_hash': None, 'config': {'config': 'value'}, 'inputs': {}, 'cache_hit': False})

        @dataclasses.dataclass
        class Outputs:
            value: typing.List[int]

        context.add_outputs('out', Outputs([1, 2, 3]))
        context.add_execution_log('name', 'task_name', '0.0.0', None, {}, {'in': OutputVariable('$output.out.value')}, False)

        logs = context.get_execution_logs()
        self.assertEqual(len(logs), 2)
        # Since the $output.out.value is a list, the value is not resolved
        self.assertEqual(logs[1], {'name': 'name', 'task_name': 'task_name', 'task_version': '0.0.0', 'task_hash': None, 'config': {}, 'inputs': {'in': '$output.out.value'}, 'cache_hit': False})


class TestExecutionLogger(unittest.TestCase):
    def test_deepcopy(self):
        execution_logger = ExecutionLogger()
        execution_logger.log('name', 'task_name', '0.0.0', None, {'config': 0}, {}, False)

        copied_logger = copy.deepcopy(execution_logger)
        copied_logger.log('name', 'task_name', '0.0.0', None, {'config': 1}, {}, False)

        execution_logger.log('name', 'task_name', '0.0.0', None, {'config': 2}, {}, False)

        logs = execution_logger.get()
        self.assertEqual(len(logs), 3)
        self.assertEqual(logs[0], {'name': 'name', 'task_name': 'task_name', 'task_version': '0.0.0', 'task_hash': None, 'config': {'config': 0}, 'inputs': {}, 'cache_hit': False})
        self.assertEqual(logs[1], {'name': 'name', 'task_name': 'task_name', 'task_version': '0.0.0', 'task_hash': None, 'config': {'config': 1}, 'inputs': {}, 'cache_hit': False})
        self.assertEqual(logs[2], {'name': 'name', 'task_name': 'task_name', 'task_version': '0.0.0', 'task_hash': None, 'config': {'config': 2}, 'inputs': {}, 'cache_hit': False})

        copied_logs = copied_logger.get()
        self.assertEqual(logs, copied_logs)



import copy
import importlib
import importlib.abc
import importlib.util
import json
import logging
import os
import pathlib
import subprocess
import sys
import tempfile
import typing

from irisml.core import JobDescription
import irisml.compiler.stubs


logger = logging.getLogger(__name__)


class Compiler:
    def compile(self, input_filepath, include_paths, build_args: typing.Dict[str, str]):
        include_paths = [str(pathlib.Path(p).resolve()) for p in include_paths]
        env = copy.deepcopy(os.environ)
        env['PYTHONPATH'] = ':'.join(include_paths) + ':' + env.get('PYTHONPATH', '')

        with tempfile.NamedTemporaryFile() as f:
            script_body = f"""
import irisml.compiler
irisml.compiler.compiler.init()
from {input_filepath.stem} import main
irisml.compiler.compiler.generate(main, "{f.name}", {repr(build_args)})
"""

            p = subprocess.run([sys.executable, '-c', script_body], capture_output=True, cwd=input_filepath.parent, env=env)

            json_str = pathlib.Path(f.name).read_text()

        stderr_str = p.stderr.decode()
        stdout_str = p.stdout.decode()

        if stdout_str:
            print(stdout_str)

        for s in stderr_str.splitlines():
            logger.warning(s)

        if p.returncode != 0:
            raise RuntimeError("Failed to compile.")

        try:
            job_description_dict = json.loads(json_str)
        except json.JSONDecodeError:
            raise RuntimeError(f"Failed to parse JSON: {stdout_str}")

        return JobDescription.from_dict(job_description_dict)


class CompilerSession:
    def __init__(self):
        self._tasks = []
        self._names = set()
        self._on_error = None

    def get_unique_name(self, task_name, name):
        """Get a unique name for the task."""
        if name:
            if name in self._names:
                raise ValueError(f'Name {name} is already used.')
            self._names.add(name)
            return name

        if task_name not in self._names:
            self._names.add(task_name)
            return task_name

        i = 1
        while True:
            name = f'{task_name}_{i}'
            if name not in self._names:
                self._names.add(name)
                return name
            i += 1

    def add_task(self, task_instance, task_name, name, inputs, config):
        implicit_name = self.get_unique_name(task_name, name)
        task = {'task': task_name, 'explicit_name': name, 'implicit_name': implicit_name, 'inputs': self._resolve_variable(inputs), 'config': self._resolve_variable(config),
                'task_instance': task_instance}
        self._tasks.append(task)
        return implicit_name

    def set_on_error(self, tasks):
        self._on_error = [t.to_dict() for t in tasks]

    def generate(self):
        def process_task(task):
            t = {'task': task['task'], 'inputs': task['inputs'], 'config': task['config']}
            if task['explicit_name']:
                t['name'] = task['explicit_name']
            elif task['task_instance'].ref_count > 0 and task['implicit_name'] != task['task']:
                t['name'] = task['implicit_name']
            return t

        return JobDescription.from_dict({'tasks': [process_task(t) for t in self._tasks], 'on_error': self._on_error})

    @staticmethod
    def _resolve_variable(data):
        if isinstance(data, irisml.compiler.stubs.VariableStub):
            return data.resolve()
        if isinstance(data, dict):
            return {k: CompilerSession._resolve_variable(v) for k, v in data.items()}
        if isinstance(data, list):
            return [CompilerSession._resolve_variable(v) for v in data]
        if isinstance(data, irisml.compiler.stubs.OutputsStub):
            raise TypeError(f"OutputsStub is not allowed in the inputs/config: {data}")
        return data


class StubImporter(importlib.abc.MetaPathFinder, importlib.abc.Loader):
    def __init__(self):
        self._enabled = True

    def find_spec(self, fullname, path, target=None):
        if fullname == 'irisml.tasks' and self._enabled:
            # Replace the tasks module with a stub only once.
            self._enabled = False
            self._original_tasks = None
            return importlib.util.spec_from_loader(fullname, self)
        return None

    def create_module(self, spec):
        return irisml.compiler.stubs.TasksModuleStub(spec.name)

    def exec_module(self, module):
        pass  # Do nothing


def init():
    """Initialize the environment for compiler."""
    sys.meta_path.insert(0, StubImporter())


def generate(main, output_filename, build_args):
    """Generate a job description from a main function."""
    session = CompilerSession()
    irisml.compiler.stubs.current_session.set(session)
    main(**build_args)
    irisml.compiler.stubs.current_session.set(None)

    job_description_dict = session.generate().to_dict()

    pathlib.Path(output_filename).write_text(json.dumps(job_description_dict, indent=4))



import argparse
import json
import pathlib
import re
from irisml.compiler import Compiler


def main():
    class KeyValuePairAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            k, v = values.split('=', 1)
            d = getattr(namespace, self.dest)
            d[k] = v

    parser = argparse.ArgumentParser()
    parser.add_argument('input_filepath', type=pathlib.Path)
    parser.add_argument('--output_filepath', '-o', type=pathlib.Path, required=True)
    parser.add_argument('-I', action='append', dest='include_paths', default=[])
    parser.add_argument('--build-arg', '-b', action=KeyValuePairAction, dest='build_args', default={})

    args = parser.parse_args()

    for key, value in args.build_args.items():
        if not re.match(r'^[a-z][a-z0-9_]*$', key):
            parser.error(f"Invalid build arg name: {key}. Must be a valid python variable name.")

    compiler = Compiler()
    job_description = compiler.compile(args.input_filepath, args.include_paths, args.build_args)
    args.output_filepath.parent.mkdir(parents=True, exist_ok=True)
    args.output_filepath.write_text(json.dumps(job_description.to_dict(), indent=4) + '\n')


if __name__ == '__main__':
    main()
